import 'dart:collection';

import 'package:flutter/foundation.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';

import 'animated_list_inervals.dart';
import 'animated_sliver_list.dart';

//---------------------------------------------------------------------------------------------
// AnimatedSliverMultiBoxAdaptorElement (Render Object Element / Child Manager)
//---------------------------------------------------------------------------------------------

/// An element that lazily builds children for a [SliverMultiBoxAdaptorWidget].
///
/// Implements [RenderSliverBoxChildManager], which lets this element manage
/// the children of subclasses of [RenderSliverMultiBoxAdaptor].
///
/// This class takes the implementation of the standard [SliverMultiBoxAdaptorElement] class
/// and provides support for animations in face of list changes and reordering.
class AnimatedSliverMultiBoxAdaptorElement extends RenderObjectElement
    implements RenderSliverBoxChildManager {
  /// Creates an element that lazily builds children for the given widget with support for
  /// animations in response to changes to the list and reordering.
  AnimatedSliverMultiBoxAdaptorElement(AnimatedSliverList widget)
      : super(widget);

  /// support for the method didChangeDependencies like in [State.didChangeDependencies]
  bool _didChangeDependencies = false;

  /// flag passed to [AnimatedSliverChildBuilderDelegate.build] method.
  bool animating;

  /// dummy element used to calculate the size of a resizing interval
  Element _dummyChild;

  /// list of updates to apply to this render object element in response to one or more changes to the list
  final List<_BuildUpdate> _updates = [];

  /// This method has been overridden in order to handle the new
  /// [AnimatedRenderSliverList.didChangeDependencies] method like in
  /// [State.didChangeDependencies].
  /// In addition the method attaches this instance to the [AnimatedListController].
  @override
  void mount(final Element parent, final dynamic newSlot) {
    super.mount(parent, newSlot);
    renderObject.didChangeDependencies(this);
    renderObject.didChildManagerMounted();
  }

  /// This method has been overridden in order to detach this instance from the
  /// [AnimatedListController] and to dispose the render object [AnimatedRenderSliverList].
  @override
  void unmount() {
    renderObject.dispose();
    renderObject.didChildManagerUnmounted();
    super.unmount();
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    _didChangeDependencies = true;
  }

  @override
  AnimatedSliverList get widget => super.widget as AnimatedSliverList;

  @override
  AnimatedRenderSliverList get renderObject =>
      super.renderObject as AnimatedRenderSliverList;

  /// Copied and adapted from [SliverMultiBoxAdaptorElement.update].
  @override
  void update(covariant AnimatedSliverList newWidget) {
    final oldWidget = widget;
    super.update(newWidget);
    final newDelegate = newWidget.delegate;
    final oldDelegate = oldWidget.delegate;
    if (newDelegate != oldDelegate &&
        (newDelegate.runtimeType != oldDelegate.runtimeType ||
            newDelegate.shouldRebuild(oldDelegate))) {
      renderObject.update(widget); // update also the render object!
      _updates.clear();
      _reorderingElement = null;
      _reorderingStartUpdate = false;
      _reorderingStopUpdate = false;
      _reorderingFromIndex = null;
      _reorderingToIndex = null;
      performRebuild();
    }
  }

  final SplayTreeMap<int, Element> _childElements =
      SplayTreeMap<int, Element>();
  RenderBox _currentBeforeChild;

  Widget _build(final int index) =>
      renderObject.buildAnimatedWidget(this, index);

  /// Copied from [SliverMultiBoxAdaptorElement.createChild].
  @override
  void createChild(final int index, {@required final RenderBox after}) {
    assert(_currentlyUpdatingChildIndex == null);
    owner.buildScope(this, () {
      final insertFirst = after == null;
      assert(insertFirst || _childElements[index - 1] != null);
      _currentBeforeChild = insertFirst
          ? null
          : (_childElements[index - 1].renderObject as RenderBox);
      Element newChild;
      try {
        _currentlyUpdatingChildIndex = index;
        newChild = updateChild(_childElements[index], _build(index), index);
      } finally {
        _currentlyUpdatingChildIndex = null;
      }
      if (newChild != null) {
        _childElements[index] = newChild;
      } else {
        _childElements.remove(index);
      }
    });
  }

  /// Copied from [SliverMultiBoxAdaptorElement.updateChild].
  @override
  Element updateChild(
      final Element child, final Widget newWidget, final dynamic newSlot) {
    final oldParentData =
        child?.renderObject?.parentData as SliverMultiBoxAdaptorParentData;
    final newChild = super.updateChild(child, newWidget, newSlot);
    final newParentData =
        newChild?.renderObject?.parentData as SliverMultiBoxAdaptorParentData;

    // Preserve the old layoutOffset if the renderObject was swapped out.
    if (oldParentData != newParentData &&
        oldParentData != null &&
        newParentData != null) {
      newParentData.layoutOffset = oldParentData.layoutOffset;
    }
    return newChild;
  }

  /// Copied from [SliverMultiBoxAdaptorElement.forgetChild].
  @override
  void forgetChild(final Element child) {
    assert(child != null);
    assert(child.slot != null);
    assert(_childElements.containsKey(child.slot));
    _childElements.remove(child.slot);
    super.forgetChild(child);
  }

  /// Copied from [SliverMultiBoxAdaptorElement.removeChild].
  @override
  void removeChild(final RenderBox child) {
    final index = renderObject.indexOf(child);
    assert(_currentlyUpdatingChildIndex == null);
    assert(index >= 0);
    owner.buildScope(this, () {
      assert(_childElements.containsKey(index));
      try {
        _currentlyUpdatingChildIndex = index;
        final result = updateChild(_childElements[index], null, index);
        assert(result == null);
      } finally {
        _currentlyUpdatingChildIndex = null;
      }
      _childElements.remove(index);
      assert(!_childElements.containsKey(index));
    });
  }

  /// Copied from [SliverMultiBoxAdaptorElement.estimateMaxScrollOffset].
  /// The [AnimatedSliverChildBuilderDelegate.estimateMaxScrollOffset] method call
  /// has been removed just now.
  @override
  double estimateMaxScrollOffset(
    final SliverConstraints constraints, {
    final int firstIndex,
    final int lastIndex,
    final double leadingScrollOffset,
    final double trailingScrollOffset,
  }) {
    final childCount = estimatedChildCount;
    if (childCount == null) return double.infinity;
    return renderObject.extrapolateMaxScrollOffset(
      firstIndex,
      lastIndex,
      leadingScrollOffset,
      trailingScrollOffset,
      childCount + renderObject.intervals.itemCountAdjustment,
    );
  }

  int get estimatedChildCount => widget.delegate.estimatedChildCount;

  /// Copied from [SliverMultiBoxAdaptorElement.childCount].
  /// The final count will be changed to take in account the last invisibile boundary item and
  /// the items taken by all intervals.
  @override
  int get childCount {
    var result = estimatedChildCount;
    if (result == null) {
      // Since childCount was called, we know that we reached the end of
      // the list (as in, _build return null once), so we know that the
      // list is finite.
      // Let's do an open-ended binary search to find the end of the list
      // manually.
      var lo = 0;
      var hi = 1;
      const max = kIsWeb
          ? 9007199254740992 // max safe integer on JS (from 0 to this number x != x+1)
          : ((1 << 63) - 1);
      while (
          renderObject.callBuildDelegateCallback(this, hi - 1, true) != null) {
        lo = hi - 1;
        if (hi < max ~/ 2) {
          hi *= 2;
        } else if (hi < max) {
          hi = max;
        } else {
          throw FlutterError(
              'Could not find the number of children in ${widget.delegate}.\n'
              'The childCount getter was called (implying that the delegate\'s builder returned null '
              'for a positive index), but even building the child with index $hi (the maximum '
              'possible integer) did not return null. Consider implementing childCount to avoid '
              'the cost of searching for the final child.');
        }
      }
      while (hi - lo > 1) {
        final mid = (hi - lo) ~/ 2 + lo;
        if (renderObject.callBuildDelegateCallback(this, mid - 1, true) ==
            null) {
          hi = mid;
        } else {
          lo = mid;
        }
      }
      result = lo;
    }
    assert(result != null);
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    return result + 1 + renderObject.intervals.itemCountAdjustment;
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
  }

  final _resizingIntervals = <AnimatedListInterval, double>{};

  /// This method is being called from [AnimatedRenderSliverList] when creating a new interval.
  /// A new [_BuildUpdate] will be added to the update list and this Element will be marked to rebuild.
  void onUpdateOnNewInterval(AnimatedListInterval interval) {
    if (interval.isInRemovingState) {
      _updates.add(_BuildUpdate(
        index: interval.index,
        removeCount: interval.removeCount,
        insertCount: interval.insertCount,
        type: _BuildUpdateType.NEW_REMOVING_INTERVAL,
      ));
    } else if (interval.isInResizingState) {
      _updates.add(_BuildUpdate(
        index: interval.index,
        removeCount: interval.removeCount,
        insertCount: interval.insertCount,
        type: _BuildUpdateType.NEW_RESIZING_INTERVAL,
      ));
      _resizingIntervals.putIfAbsent(interval, () => interval.fromSize);
    }
    markNeedsBuild();
  }

  /// This method is being called from [AnimatedRenderSliverList] when a removing interval becomes
  /// a resizing interval.
  /// A new [_BuildUpdate] will be added to the update list and this Element will be marked to rebuild.
  void onUpdateOnIntervalRemovedToResizing(AnimatedListInterval interval) {
    _updates.add(_BuildUpdate(
      index: interval.index,
      removeCount: interval.removeCount,
      insertCount: interval.insertCount,
      type: _BuildUpdateType.REMOVED_TO_RESIZING,
    ));
    _resizingIntervals.putIfAbsent(interval, () => interval.fromSize);
    markNeedsBuild();
  }

  /// This method is being called from [AnimatedRenderSliverList] when a resizing interval becomes
  /// an inserting interval.
  /// A new [_BuildUpdate] will be added to the update list and this Element will be marked to rebuild.
  void onUpdateOnIntervalResizedToInserting(AnimatedListInterval interval) {
    _updates.add(_BuildUpdate(
      index: interval.index,
      removeCount: interval.removeCount,
      insertCount: interval.insertCount,
      type: _BuildUpdateType.RESIZED_TO_INSERTING,
    ));
    markNeedsBuild();
  }

  /// This method is being called from [AnimatedRenderSliverList] when a resizing interval (without
  /// insertion) completes.
  /// A new [_BuildUpdate] will be added to the update list and this Element will be marked to rebuild.
  void onUpdateOnIntervalResizedToDisposing(AnimatedListInterval interval) {
    _updates.add(_BuildUpdate(
      index: interval.index,
      removeCount: interval.removeCount,
      insertCount: interval.insertCount,
      type: _BuildUpdateType.RESIZED_TO_DISPOSING,
    ));
    markNeedsBuild();
  }

  /// This method is being called from [AnimatedRenderSliverList] when a inserting interval completes.
  /// This Element will be marked to rebuild.
  void onUpdateOnIntervalInsertedToDisposing(AnimatedListInterval interval) {
    markNeedsBuild();
  }

  /// This method is being called from [AnimatedRenderSliverList] when a changing interval completes.
  /// This Element will be marked to rebuild.
  void onUpdateOnIntervalChangedToDisposing(AnimatedListInterval interval) {
    markNeedsBuild();
  }

  Element _reorderingElement;
  bool _reorderingStartUpdate = false, _reorderingStopUpdate = false;
  int _reorderingFromIndex, _reorderingToIndex;

  RenderBox get reorderingDraggedChild =>
      _reorderingElement?.renderObject as RenderBox;

  RenderBox onUpdateOnStartReording(int index) {
    assert(index != null);
    markNeedsBuild();
    _reorderingStartUpdate = true;
    _reorderingFromIndex = index;
    return _childElements[index].renderObject as RenderBox;
  }

  void onUpdateOnStopReording(int fromIndex, int toIndex) {
    markNeedsBuild();
    assert(toIndex != null && _reorderingFromIndex == fromIndex);
    _reorderingStopUpdate = true;
    _reorderingToIndex = toIndex;
    _resizingIntervals.clear();
  }

  double calculateOffsetCorrection(int firstIndex) {
    var ret = 0.0;
    for (var entry in _resizingIntervals.entries) {
      final interval = entry.key;
      final lastsize = entry.value ?? interval.fromSize;
      final currentSize =
          interval.isInResizingState ? interval.currentSize : interval.toSize;
      if (interval.index < firstIndex) {
        ret += currentSize - lastsize;
      }
      if (interval.isInResizingState) {
        _resizingIntervals[interval] = currentSize;
      }
    }
    _resizingIntervals
        .removeWhere((interval, lastSize) => !interval.isInResizingState);
    return ret;
  }

  /// Copied and adapted from [SliverMultiBoxAdaptorElement.performRebuild].
  /// In addition:
  /// - calls the [AnimatedRenderSliverList.didChangeDependencies] method if
  ///   a dependency has been changed;
  /// - considers all build updates [_BuildUpdate] in the list to move the old
  ///   rendered children to the correct new positions;
  /// - on start reordering it cretes a new off-list dragged child;
  /// - on stop reordering it destroys the off-list dragged child and repositions
  ///   the children taking into account the swap;
  @override
  void performRebuild() {
    // var qqqq = List.from(_updates);
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    if (_didChangeDependencies) {
      renderObject.didChangeDependencies(this);
      _didChangeDependencies = false;
    }
    super.performRebuild();
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    _currentBeforeChild = null;
    assert(_currentlyUpdatingChildIndex == null);
    try {
      final newChildren = SplayTreeMap<int, Element>();
      final removedChildren = <int>{};
      final Map<int, double> indexToLayoutOffset = HashMap<int, double>();

      if (_reorderingStartUpdate) {
        _createReorderDraggedChild();
      }

      if (_reorderingStopUpdate) {
        _destroyExternalChild(_reorderingElement);
        _reorderingElement = null;
      }

      for (final index in _childElements.keys.toList()) {
        assert(_childElements[index] != null);

        //""""""""""""""""""""""""""""""""""""""""""""""""""""""
        // final key = _childElements[index].widget.key;
        // var newIndex = key == null ? null : widget.delegate.findIndexByKey(key);
        int newIndex;
        newIndex = _oldIndexToNewIndex(index);
        if (_reorderingStopUpdate && newIndex != null) {
          if (_reorderingFromIndex < _reorderingToIndex) {
            if (newIndex >= _reorderingFromIndex) {
              if (newIndex == _reorderingFromIndex) {
                newIndex = _reorderingToIndex;
              } else if (newIndex <= _reorderingToIndex) {
                newIndex--;
              }
            }
          } else if (_reorderingFromIndex > _reorderingToIndex) {
            if (newIndex >= _reorderingToIndex) {
              if (newIndex == _reorderingFromIndex) {
                newIndex = _reorderingToIndex;
              } else if (newIndex <= _reorderingFromIndex) {
                newIndex++;
              }
            }
          }
        }
        //""""""""""""""""""""""""""""""""""""""""""""""""""""""

        final childParentData = _childElements[index].renderObject?.parentData
            as SliverMultiBoxAdaptorParentData;

        if (childParentData != null && childParentData.layoutOffset != null) {
          // indexToLayoutOffset[index] = childParentData.layoutOffset;
          indexToLayoutOffset[newIndex] = childParentData.layoutOffset;
        }

        if (newIndex != null && newIndex != index) {
          // The layout offset of the child being moved is no longer accurate.
          if (childParentData != null) {
            childParentData.layoutOffset = null;
          }

          newChildren[newIndex] = _childElements[index];
          // We need to make sure the original index gets processed.
          newChildren.putIfAbsent(index, () => null);
          // We do not want the remapped child to get deactivated during processElement.
          _childElements.remove(index);
        } else {
          // newChildren.putIfAbsent(index, () => _childElements[index]);
          newChildren.putIfAbsent(index,
              newIndex != null ? () => _childElements[index] : () => null);
        }

        // remove all kept alive children that are about to become animated first
        if (newChildren[index] != null &&
            newIndex != null &&
            renderObject.isIndexAnimated(newIndex) &&
            (newChildren[index].renderObject.parentData
                    as SliverMultiBoxAdaptorParentData)
                .keptAlive) {
          _currentlyUpdatingChildIndex = index;
          updateChild(newChildren[newIndex], null, index);
          newChildren[newIndex] = null;
          removedChildren.add(newIndex);
        }
      }

      void processElement(int index) {
        _currentlyUpdatingChildIndex = index;
        if (_childElements[index] != null &&
            _childElements[index] != newChildren[index] &&
            !removedChildren.contains(index)) {
          _childElements[index] =
              updateChild(_childElements[index], null, index);
        }

        final newChild = updateChild(newChildren[index], _build(index), index);
        if (newChild != null) {
          _childElements[index] = newChild;
          final parentData = newChild.renderObject.parentData
              as SliverMultiBoxAdaptorParentData;
          if (index == 0) {
            parentData.layoutOffset = 0.0;
          } else if (indexToLayoutOffset.containsKey(index)) {
            parentData.layoutOffset = indexToLayoutOffset[index];
          }
          if (!parentData.keptAlive) {
            _currentBeforeChild = newChild.renderObject as RenderBox;
          }
        } else {
          _childElements.remove(index);
        }
      } // processElement

      // Moving children will temporary violate the integrity.
      renderObject.debugChildIntegrityEnabled = false;
      newChildren.keys.forEach(processElement); // !!

      if (_didUnderflow) {
        final lastKey = _childElements.lastKey() ?? -1;
        final rightBoundary = lastKey + 1;
        newChildren[rightBoundary] = _childElements[rightBoundary];
        processElement(rightBoundary);
      }

      assert(() {
        int j;
        _childElements.keys.forEach((i) {
          if (renderObject.isBoundaryChild(
              _childElements[i].renderObject as RenderBox)) j = i;
        });
        return j == null || j == _childElements.lastKey();
      }());
    } finally {
      _currentlyUpdatingChildIndex = null;
      _updates.clear();
      _reorderingStartUpdate = false;
      _reorderingStopUpdate = false;
    }
  }

  void _createReorderDraggedChild() {
    _currentlyUpdatingChildIndex = null;
    var dragWidget = renderObject.buildDraggedChild(this, _reorderingFromIndex);
    _reorderingElement = updateChild(_reorderingElement, dragWidget, null);
  }

  void _destroyExternalChild(Element e) {
    _currentlyUpdatingChildIndex = e.slot as int;
    (e.renderObject.parentData as SliverMultiBoxAdaptorParentData).index = null;
    updateChild(e, null, e.slot as int);
  }

  int _oldIndexToNewIndex(int index) {
    for (final upd in _updates) {
      switch (upd.type) {
        case _BuildUpdateType.NEW_RESIZING_INTERVAL:
          if (index >= upd.index) index++;
          break;
        case _BuildUpdateType.REMOVED_TO_RESIZING:
          if (index >= upd.index && index < upd.index + upd.removeCount) {
            return null;
          }
          if (index >= upd.index) index += 1 - upd.removeCount;
          break;
        case _BuildUpdateType.RESIZED_TO_INSERTING:
          if (index == upd.index) return null;
          if (index > upd.index) index += upd.insertCount - 1;
          break;
        case _BuildUpdateType.RESIZED_TO_DISPOSING:
          if (index == upd.index) return null;
          if (index > upd.index) index--;
          break;
        default:
          break;
      }
    }
    return index;
  }

  int _currentlyUpdatingChildIndex;

  /// Copied from [SliverMultiBoxAdaptorElement.debugAssertChildListLocked].
  @override
  bool debugAssertChildListLocked() {
    assert(_currentlyUpdatingChildIndex == null);
    return true;
  }

  /// Copied from [SliverMultiBoxAdaptorElement.didAdoptChild].
  /// An assertion was removed to accept the dummy child element.
  @override
  void didAdoptChild(final RenderBox child) {
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    // assert(_currentlyUpdatingChildIndex != null);
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    final childParentData = child.parentData as SliverMultiBoxAdaptorParentData;
    childParentData.index = _currentlyUpdatingChildIndex;
  }

  bool _didUnderflow = false;

  /// Copied from [SliverMultiBoxAdaptorElement.debugAssertChildListLocked].
  @override
  void setDidUnderflow(final bool value) {
    _didUnderflow = value;
  }

  /// Copied from [SliverMultiBoxAdaptorElement.insertRenderObjectChild].
  /// This method has been changed in order to accept a null slot, in case of
  /// dummy child element.
  @override
  void insertRenderObjectChild(covariant RenderObject child, final int slot) {
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    if (slot == null) {
      renderObject.setupParentData(child);
      assert(renderObject.debugValidateChild(child));
      renderObject.adoptChild(child);
      return;
    }
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""

    assert(slot != null);
    assert(_currentlyUpdatingChildIndex == slot);
    assert(renderObject.debugValidateChild(child));
    renderObject.insert(child as RenderBox, after: _currentBeforeChild);
    assert(() {
      final childParentData =
          child.parentData as SliverMultiBoxAdaptorParentData;
      assert(slot == childParentData.index);
      return true;
    }());
  }

  /// Copied from [SliverMultiBoxAdaptorElement.moveRenderObjectChild].
  @override
  void moveRenderObjectChild(
      covariant RenderObject child, final int oldSlot, final int newSlot) {
    assert(newSlot != null);
    assert(_currentlyUpdatingChildIndex == newSlot);
    renderObject.move(child as RenderBox, after: _currentBeforeChild);
  }

  /// Copied from [SliverMultiBoxAdaptorElement.removeRenderObjectChild].
  /// This method has been changed to remove the dummy child element correctly.
  @override
  void removeRenderObjectChild(covariant RenderObject child, final int slot) {
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    if ((child.parentData as SliverMultiBoxAdaptorParentData).index == null) {
      renderObject.dropChild(child);
      return;
    }
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""

    assert(_currentlyUpdatingChildIndex != null);
    renderObject.remove(child as RenderBox);
  }

  /// Copied from [SliverMultiBoxAdaptorElement.visitChildren].
  @override
  void visitChildren(final ElementVisitor visitor) {
    // The toList() is to make a copy so that the underlying list can be modified by
    // the visitor:
    assert(!_childElements.values.any((Element child) => child == null));
    _childElements.values.cast<Element>().toList().forEach(visitor);
    // if (_reorderingChild != null) visitor.call(_reorderingChild);
    if (_reorderingElement != null) {
      visitor.call(_reorderingElement);
    }
  }

  /// Copied from [SliverMultiBoxAdaptorElement.debugVisitOnstageChildren].
  @override
  void debugVisitOnstageChildren(final ElementVisitor visitor) {
    _childElements.values.cast<Element>().where((Element child) {
      final parentData =
          child.renderObject.parentData as SliverMultiBoxAdaptorParentData;
      double itemExtent;
      switch (renderObject.constraints.axis) {
        case Axis.horizontal:
          itemExtent = child.renderObject.paintBounds.width;
          break;
        case Axis.vertical:
          itemExtent = child.renderObject.paintBounds.height;
          break;
      }

      return parentData.layoutOffset != null &&
          parentData.layoutOffset <
              renderObject.constraints.scrollOffset +
                  renderObject.constraints.remainingPaintExtent &&
          parentData.layoutOffset + itemExtent >
              renderObject.constraints.scrollOffset;
    }).forEach(visitor);
  }

  /// Layouts the dummy child element.
  /// This method is called by [AnimatedRenderSliverList._measureIntervalSize]
  /// when calculating the size of a resizing interval.
  RenderBox layoutDummyChild(
      final Widget widget, final BoxConstraints childConstraints) {
    owner.buildScope(this, () {
      _currentlyUpdatingChildIndex = null;
      _dummyChild = updateChild(_dummyChild, widget, null);
      _dummyChild.renderObject.layout(childConstraints, parentUsesSize: true);
    });
    return _dummyChild.renderObject as RenderBox;
  }

  /// Frees the dummy child element.
  /// This method is called by [AnimatedRenderSliverList._measureIntervalSize]
  /// when calculating the size of a resizing interval.
  void dropDummyChild() {
    if (_dummyChild != null) {
      owner.buildScope(this, () {
        updateChild(_dummyChild, null, null);
        _dummyChild = null;
      });
    }
  }

  /// Copied from [SliverMultiBoxAdaptorElement.didStartLayout].
  @override
  void didStartLayout() {
    assert(debugAssertChildListLocked());
  }

  /// Copied from [SliverMultiBoxAdaptorElement.didFinishLayout].
  @override
  void didFinishLayout() {
    assert(debugAssertChildListLocked());
    final firstIndex = _childElements.firstKey() ?? 0;
    final lastIndex = _childElements.lastKey() ?? 0;
    widget.delegate.didFinishLayout(firstIndex, lastIndex);
  }

  AnimatedListInterval intervalAtIndex(int index) {
    var key = _childElements[index]?.widget?.key;
    if (key is ValueKey && key.value is AnimatedListInterval) {
      return key.value as AnimatedListInterval;
    }
    return null;
  }

  Widget widgetOf(int index) =>
      index == null ? null : _childElements[index]?.widget;

  bool get havePendingUpdates => _updates.isNotEmpty;
}

enum _BuildUpdateType {
  NEW_REMOVING_INTERVAL,
  NEW_RESIZING_INTERVAL,
  REMOVED_TO_RESIZING,
  RESIZED_TO_INSERTING,
  RESIZED_TO_DISPOSING,
}

class _BuildUpdate {
  final int index;
  final int removeCount;
  final int insertCount;
  final _BuildUpdateType type;

  _BuildUpdate({this.index, this.removeCount, this.insertCount, this.type});
}
