import 'dart:math';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/scheduler.dart';
import 'package:flutter/widgets.dart';

import 'animated_list_child_manager.dart';
import 'animated_list_inervals.dart';
import 'morph_transition.dart';

const _kAnimationSpeed = 500;

const _kReorderAnimationSpeed = 250;

const _kReorderingScrollSpeed = 10.0;

const Duration _kRemovingDuration = Duration(milliseconds: _kAnimationSpeed);
const Duration _kResizingDuration = Duration(milliseconds: _kAnimationSpeed);
const Duration _kInsertingDuration = Duration(milliseconds: _kAnimationSpeed);
const Duration _kChangingDuration = Duration(milliseconds: _kAnimationSpeed);
const Duration _kReorderingDuration =
    Duration(milliseconds: _kReorderAnimationSpeed);

const Curve _kRemovingCurve = Curves.easeOut;
const Curve _kResizingCurve = Curves.easeInOut;
const Curve _kInsertingCurve = Curves.easeIn;
const Curve _kChangingCurve = Curves.linear;
const Curve _kReorderingCurve = Curves.easeInOut;

/// Provide durations and curves for all animations of the [AnimatedSliverList].
class AnimatedListAnimationSettings {
  final Duration removingDuration,
      resizingDuration,
      insertingDuration,
      changingDuration,
      reorderingDuration;
  final Curve removingCurve,
      resizingCurve,
      insertingCurve,
      changingCurve,
      reorderingCurve;

  const AnimatedListAnimationSettings({
    this.removingDuration = _kRemovingDuration,
    this.resizingDuration = _kResizingDuration,
    this.insertingDuration = _kInsertingDuration,
    this.changingDuration = _kChangingDuration,
    this.reorderingDuration = _kReorderingDuration,
    this.removingCurve = _kRemovingCurve,
    this.resizingCurve = _kResizingCurve,
    this.insertingCurve = _kInsertingCurve,
    this.changingCurve = _kChangingCurve,
    this.reorderingCurve = _kReorderingCurve,
  });
}

//---------------------------------------------------------------------------------------------
// AnimatedListController
//---------------------------------------------------------------------------------------------

/// Use this controller to dispatch the changes of your list to the [AnimatedSliverList].
class AnimatedListController {
  AnimatedRenderSliverList _adaptor;

  AnimatedRenderSliverList get adaptor => _adaptor;

  void _checkAdaptor() {
    if (_adaptor == null) {
      throw 'This controller has not been assigned to an AnimatedListSliver widget yet.';
    }
  }

  void _setAdaptor(AnimatedRenderSliverList adaptor) {
    if (_adaptor == adaptor) return;
    if (_adaptor != null) {
      throw 'This controller cannot be assigned to multiple AnimatedListSliver widgets.';
    }
    _adaptor = adaptor;
  }

  /// Notify the [AnimatedSliverList] that a new range starting from [from] and [count] long
  /// has been inserted. Call this method after actually you have updated your list.
  void notifyInsertedRange(final int from, final int count) {
    _checkAdaptor();

    if (_adaptor.isReordering) {
      _adaptor._stopReordering(true);
      WidgetsBinding.instance.addPostFrameCallback((_) {
        notifyInsertedRange(from, count);
      });
      return;
    }

    _adaptor._notifyReplacedRange(from, 0, count, null);
  }

  /// Notify the [AnimatedSliverList] that a range starting from [from] and [count] long
  /// has been removed. Call this method after actually you have updated your list.
  /// A new builder [removeItemBuilder] has to be provided in order to build the removed
  /// items when animating.
  void notifyRemovedRange(final int from, final int count,
      final NullableIndexedWidgetBuilder removeItemBuilder) {
    _checkAdaptor();

    if (_adaptor.isReordering) {
      _adaptor._stopReordering(true);
      WidgetsBinding.instance.addPostFrameCallback((_) {
        notifyRemovedRange(from, count, removeItemBuilder);
      });
      return;
    }

    _adaptor._notifyReplacedRange(from, count, 0, removeItemBuilder);
  }

  /// Notify the [AnimatedSliverList] that a range starting from [from] and [removeCount] long
  /// has been replaced with a new [insertCount] long range. Call this method after
  /// you have updated your list.
  /// A new builder [removeItemBuilder] has to be provided in order to build the replaced
  /// items when animating.
  void notifyReplacedRange(
      final int from,
      final int removeCount,
      final int insertCount,
      final NullableIndexedWidgetBuilder removeItemBuilder) {
    _checkAdaptor();

    if (_adaptor.isReordering) {
      _adaptor._stopReordering(true);
      WidgetsBinding.instance.addPostFrameCallback((_) {
        notifyReplacedRange(from, removeCount, insertCount, removeItemBuilder);
      });
      return;
    }

    _adaptor._notifyReplacedRange(
        from, removeCount, insertCount, removeItemBuilder);
  }

  /// Notify the [AnimatedSliverList] that a range starting from [from] and [count] long
  /// has been modified. Call this method after actually you have updated your list.
  /// A new builder [changeItemBuilder] has to be provided in order to build the old
  /// items when animating.
  void notifyChangedRange(final int from, final int count,
      final NullableIndexedWidgetBuilder changeItemBuilder) {
    _checkAdaptor();

    if (_adaptor.isReordering) {
      _adaptor._stopReordering(true);
      WidgetsBinding.instance.addPostFrameCallback((_) {
        notifyChangedRange(from, count, changeItemBuilder);
      });
      return;
    }

    _adaptor._notifyChangedRange(from, count, changeItemBuilder);
  }

  /// After notifying your changes, call this method in order to dispatch them to the
  /// [AnimatedSliverList]. Eventually the animations that follow will start.
  void dispatchChanges() {
    _checkAdaptor();

    if (_adaptor.isReordering) {
      _adaptor._stopReordering(true);
      WidgetsBinding.instance.addPostFrameCallback((_) {
        dispatchChanges();
      });
      return;
    }

    _adaptor._dispatchChanges();
  }
}

//---------------------------------------------------------------------------------------------
// AnimatedListAnimationBuilder
//---------------------------------------------------------------------------------------------

/// Abstract class for building animated Widgets.
abstract class AnimatedListAnimationBuilder {
  /// Wraps the [widget] in an animated widget for removing items.
  Widget buildRemoving(
      BuildContext context, Widget widget, Animation<double> animation);

  /// Wraps the [widget] in an animated widget for inserting new items.
  Widget buildInserting(
      BuildContext context, Widget widget, Animation<double> animation);

  /// Wraps the old Widget [fromWidget] and to new Widget [toWidget] in an animated one for changing items.
  Widget buildChanging(BuildContext context, Widget fromWidget, Widget toWidget,
      Animation<double> animation);

  /// Wraps the [widget] in a dragged widget during reordering.
  Widget buildReordering(BuildContext context, Widget widget, int index);
}

/// Default singleton class for building animated Widgets.
///
/// For removing items a fade out effect will be provided.
/// For inserting new items a fade in effect will be provided.
/// For changing items a morph effect will be provided.
/// For dragged items an opacity effect of 75% will be provided.
class DefaultAnimatedListAnimationBuilder
    implements AnimatedListAnimationBuilder {
  const DefaultAnimatedListAnimationBuilder._();

  static const DefaultAnimatedListAnimationBuilder instance =
      DefaultAnimatedListAnimationBuilder._();

  @override
  Widget buildRemoving(
      BuildContext context, Widget widget, Animation<double> animation) {
    return FadeTransition(
        child: widget, opacity: Tween(begin: 1.0, end: 0.0).animate(animation));
  }

  @override
  Widget buildInserting(
      BuildContext context, Widget widget, Animation<double> animation) {
    return FadeTransition(child: widget, opacity: animation);
  }

  @override
  Widget buildChanging(BuildContext context, Widget fromWidget, Widget toWidget,
      Animation<double> animation) {
    return MorphTransition(
        fromChild: fromWidget, toChild: toWidget, animation: animation);
  }

  @override
  Widget buildReordering(BuildContext context, Widget widget, int index) {
    return Opacity(
        child: Material(elevation: 16.0, child: widget), opacity: 0.8);
  }
}

//---------------------------------------------------------------------------------------------
// AnimatedSliverList (Widget)
//---------------------------------------------------------------------------------------------

/// An animated version of the standard [SliverList].
///
/// A new delegate version of [SliverChildBuilderDelegate] called [AnimatedSliverChildBuilderDelegate]
/// has to be provided.
///
/// The [AnimatedListController] controller should also be  provided to notify this sliver
/// when a change in the underlying list occurs.
///
/// A custom [AnimatedListAnimationBuilder] can be also provided to customize animations.
///
/// A [AnimatedListAnimationSettings] can be also provided to customize durations and curves of the animations.
///
/// The [updateScrollableWhenResizing] attribute if is set to false prevents scroll listeners for being
/// notified when a resize animation occurs.
///
/// The [coordinateAnimations] attribute if is set to false to prevents coordination of animations. If true
/// (as default) all animations will follow some priorities: intervals in removal state will start first,
/// followed by those in resizing and changing state, and finally those in inserting state.
///
/// Set the [reorderable] attribute to true to enable to reordering of the list, by dragging the
/// item to reorder after long touching it.
class AnimatedSliverList extends SliverWithKeepAliveWidget {
  final AnimatedSliverChildBuilderDelegate delegate;
  final AnimatedListController controller;
  final AnimatedListAnimationBuilder animationBuilder;
  final AnimatedListAnimationSettings animatedListAnimationSettings;
  final bool updateScrollableWhenResizing;
  final bool coordinateAnimations;
  final bool reorderable;

  const AnimatedSliverList({
    Key key,
    this.delegate,
    this.controller,
    this.animationBuilder = DefaultAnimatedListAnimationBuilder.instance,
    this.animatedListAnimationSettings = const AnimatedListAnimationSettings(),
    this.updateScrollableWhenResizing = true,
    this.coordinateAnimations = true,
    this.reorderable = false,
  }) : super(key: key);

  @override
  AnimatedSliverMultiBoxAdaptorElement createElement() =>
      AnimatedSliverMultiBoxAdaptorElement(this);

  @override
  AnimatedRenderSliverList createRenderObject(BuildContext context) {
    return AnimatedRenderSliverList(
        childManager: context as AnimatedSliverMultiBoxAdaptorElement,
        widget: this);
  }
}

//---------------------------------------------------------------------------------------------
// AnimatedRenderSliverList (Render Object)
//---------------------------------------------------------------------------------------------

/// Render class of the [AnimatedSliverList].
/// This is an extension of the original class [RenderSliverList].
class AnimatedRenderSliverList extends RenderSliverList
    implements TickerProvider {
  AnimatedSliverList get widget => childManager.widget;

  final AnimatedListIntervalList intervals = AnimatedListIntervalList();
  Set<_WidgetTicker> _tickers;

  /// Creates a sliver that places multiple box children in a linear array along
  /// the main axis, taking in account of animations resulting from changes
  /// to the underlying list.
  /// Those children can also be reordered.
  AnimatedRenderSliverList({
    @required AnimatedSliverMultiBoxAdaptorElement childManager,
    AnimatedSliverList widget,
  }) : super(childManager: childManager);

  /// Updates this render object against a new [AnimatedSliverList] Widget.
  void update(AnimatedSliverList widget) {
    widget.controller._setAdaptor(this);
    intervals.clear();
    _notifying = false;
    _clearReordering();
    _tickers?.forEach((t) => _removeTicker(t));
    _tickers?.clear();
  }

  /// Creates a new ticker for animations.
  @override
  Ticker createTicker(final TickerCallback onTick) {
    var onTickNew = (Duration elapsed) {
      _markNeedsLayoutIfResizing();
      onTick(elapsed);
    };
    _tickers ??= <_WidgetTicker>{};
    final result = _WidgetTicker(onTickNew, this);
    _tickers.add(result);
    return result;
  }

  /// Marks this render object to relayout its children when animating resizing or
  /// changing intervals.
  void _markNeedsLayoutIfResizing() {
    for (final interval in intervals) {
      if (interval.isInResizingState || interval.isInChangingState) {
        markNeedsLayout();
        return;
      }
    }
  }

  /// Removes a ticker.
  void _removeTicker(_WidgetTicker ticker) {
    assert(_tickers != null);
    assert(_tickers.contains(ticker));
    _tickers.remove(ticker);
  }

  /// Disposing method that clears all previously created tickers.
  void dispose() {
    intervals.clear();
    assert(() {
      if (_tickers != null) {
        for (final ticker in _tickers) {
          if (ticker.isActive) {
            throw '$this was disposed with an active Ticker.';
          }
        }
      }
      return true;
    }());
  }

  /// A [didChangeDependencies] method like in [State.didChangeDependencies],
  /// necessary to update the tickers correctly when the "muted" attribute changes.
  void didChangeDependencies(BuildContext context) {
    final muted = !TickerMode.of(context);
    if (_tickers != null) {
      for (final ticker in _tickers) {
        ticker.muted = muted;
      }
    }
  }

  /// This method is responsible for creating the dragged Widget during reordering.
  Widget buildDraggedChild(BuildContext context, int index) {
    return widget.animationBuilder.buildReordering(
        context, widget.delegate.builder(context, index, false), index);
  }

  /// This method is responsible for creating the correct Widget for a given child's index.
  Widget buildAnimatedWidget(BuildContext context, int index) {
    var adj = 0;
    for (var j = 0; j < intervals.length; j++) {
      var interval = intervals[j];

      switch (interval.state) {
        case AnimatedListIntervalState.REMOVING:
          if (index >= interval.index) {
            if (index < (interval.index + interval.removeCount)) {
              return AbsorbPointer(
                  child: widget.animationBuilder.buildRemoving(
                context,
                interval.removeItemBuilder(context, index - interval.index),
                interval.animation,
              ));
            } else {
              adj += interval.insertCount - interval.removeCount;
            }
          }
          break;
        case AnimatedListIntervalState.CHANGING:
          if (index >= interval.index) {
            if (index < (interval.index + interval.removeCount)) {
              return AbsorbPointer(
                  child: widget.animationBuilder.buildChanging(
                context,
                interval.removeItemBuilder(context, index - interval.index),
                _defaultInsertItemBuilder(
                    context, index - interval.index, interval),
                interval.animation,
              ));
            } else {
              adj += interval.insertCount - interval.removeCount;
            }
          }
          break;
        case AnimatedListIntervalState.RESIZING:
          if (index == interval.index) {
            return Container(key: ValueKey(interval));
          } else if (index > interval.index) {
            adj += interval.insertCount - 1;
          }
          break;
        case AnimatedListIntervalState.INSERTING:
          if (index >= interval.index &&
              index < (interval.index + interval.insertCount)) {
            return AbsorbPointer(
                child: widget.animationBuilder.buildInserting(
              context,
              _defaultInsertItemBuilder(
                  context, index - interval.index, interval),
              interval.animation,
            ));
          }
          break;
        default:
          throw 'Illegal state';
      }
    }
    index += adj;

    /// Create a still (not animating) widget
    var w = callBuildDelegateCallback(context, index, false);

    // we insert a invisible boundary item at the end of the list in order to have
    // always an item in the viewport on the borders between this sliver and the next, so
    // to avoid a recalculation of the items in performLayout from scratch, thus causing
    // an annoying scroll jump due to an estimation error in calculating the size of
    // one or more resizing intervals
    if (w == null &&
        callBuildDelegateCallback(context, index - 1, true) != null) {
      return _BoundaryChild();
    }

    return w;
  }

  /// Delegates the creation of the item Widget to the
  /// [AnimatedSliverChildBuilderDelegate.build] method.
  Widget callBuildDelegateCallback(
      BuildContext context, int index, bool animating) {
    childManager.animating = animating;
    return widget.delegate.build(context, index);
  }

  bool _notifying = false;

  /// Called from [AnimatedListController] when the user notified that a range of its list has been
  /// replaced by a new one. This method is also called in case of a simple insertion or removal.
  void _notifyReplacedRange(
      final int from,
      final int removeCount,
      final int insertCount,
      final NullableIndexedWidgetBuilder removeItemBuilder) {
    if (removeCount == 0 && insertCount == 0) return;
    _notifying = true;
    _insertAndAdjust(from, removeCount, insertCount, removeItemBuilder,
        _insertNewInterval, false);
  }

  /// Called from [AnimatedListController] when the user notified that a range of its list has been
  /// changed.
  void _notifyChangedRange(final int from, final int count,
      final NullableIndexedWidgetBuilder changeItemBuilder) {
    if (count == 0) return;
    _notifying = true;
    _insertAndAdjust(from, count, count, changeItemBuilder,
        _insertNewChangingInterval, true);
  }

  /// Called from [AnimatedListController] when the user has finished notifying any changes of its list.
  /// All pending notification will be optimized first, then all animations that follow will start.
  void _dispatchChanges() {
    intervals.optimize();
    _notifying = false;
    _startAnimations();
  }

  /// Core method for adapting or creating new intervals on a new user notification.
  void _insertAndAdjust(
    int from,
    int removeCount,
    int insertCount,
    final NullableIndexedWidgetBuilder removeItemBuilder,
    final AnimatedListIntervalCreationCallback callback,
    final bool animating,
  ) {
    var adjust = 0;
    var rba = 0;
    for (var i = 0; i < intervals.length; i++) {
      final to = from + removeCount;

      final interval = intervals[i];
      final ifrom = interval.index + adjust;
      final ito = ifrom + interval.insertCount;

      //  from       to                          ifrom       ito
      //   v          v                            v          v
      //   +++++++++++  (new rem/chg interval)     ooooooooooo  (existing final (after inserting) interval)
      // Note: X indicates overlap

      int adjustInterval(
          int rem, int ins, int leading, int trailing, int rbaAdjust) {
        return intervals.adjustInterval(
            interval,
            rem,
            ins,
            leading,
            trailing,
            widget.animationBuilder,
            _shiftItemBuilder(removeItemBuilder, rba + rbaAdjust),
            callback,
            animating);
      }

      if (from < ifrom) {
        if (to <= ifrom) {
          // [+++  ooooo] -> [+++++ooooo]
          break;
        } else if (to <= ito) {
          // [+++++XXXoo] -> [+++++XXXXX]
          final leading = ifrom - from;
          final rem = removeCount - leading;
          final ins = min(rem, insertCount);
          adjustInterval(rem, ins, 0, ito - to, leading);
          insertCount -= ins;
          removeCount -= rem;
          break;
        } else {
          // [+++++XXXXX+++++]
          final overlapSize = ito - ifrom;
          final leading = ifrom - from;
          var rem = leading;
          var ins = min(rem, insertCount);
          callback(from - adjust, rem, ins,
              _shiftItemBuilder(removeItemBuilder, rba));
          i++;
          rba += leading;
          insertCount -= ins;
          removeCount -= rem;
          rem = overlapSize;
          ins = min(rem, insertCount);
          final adj = interval.adjustingItemCount;
          final a = adjustInterval(rem, ins, 0, 0, 0);
          if (a >= 0) {
            adjust += adj;
            from = ito;
          }
          i += a;
          rba += overlapSize;
          insertCount -= ins;
          removeCount -= rem;
          continue;
        }
      } else if (from < ito) {
        if (to <= ito) {
          // [ooXXoo] -> [XXXXXX]
          adjustInterval(removeCount, insertCount, from - ifrom, ito - to, 0);
          return;
        } else {
          // [ooXXXX++++++] -> [XXXXXX++++++]
          final rem = ito - from;
          final ins = min(rem, insertCount);
          final adj = interval.adjustingItemCount;
          final a = adjustInterval(rem, ins, from - ifrom, 0, 0);
          if (a >= 0) {
            adjust += adj;
            from = ito;
          }
          i += a;
          rba += rem;
          insertCount -= ins;
          removeCount -= rem;
          continue;
        }
      } else {
        // [oooooo++++++++] -> [oooooo   ++++++++]
      }
      adjust += interval.adjustingItemCount;
    }

    callback(from - adjust, removeCount, insertCount,
        _shiftItemBuilder(removeItemBuilder, rba));
  }

  /// Creates a new interval in initial removing or resizing status.
  AnimatedListInterval _insertNewInterval(
      final int from,
      final int removeCount,
      final int insertCount,
      final NullableIndexedWidgetBuilder removeItemBuilder,
      [final AnimatedListIntervalEventCallback onDisposed]) {
    if (removeCount == 0 && insertCount == 0) return null;
    var interval = intervals.insertReplacingInterval(
      vsync: this,
      animationSettings: widget.animatedListAnimationSettings,
      index: from,
      insertCount: insertCount,
      removeCount: removeCount,
      removeItemBuilder: removeItemBuilder,
      onRemovingCompleted: _onIntervalRemovingCompleted,
      onResizingCompleted: _onIntervalResizingCompleted,
      onInsertingCompleted: _onIntervalInsertingCompleted,
      onCompleted: _onIntervalCompleted,
      onDisposed: onDisposed,
    );
    childManager.onUpdateOnNewInterval(interval);
    return interval;
  }

  /// Creates a new interval in initial changing status.
  AnimatedListInterval _insertNewChangingInterval(
      final int from,
      final int removeCount,
      final int insertCount,
      final NullableIndexedWidgetBuilder removeItemBuilder,
      [final AnimatedListIntervalEventCallback onDisposed]) {
    assert(removeCount == insertCount);
    if (removeCount == 0 && insertCount == 0) return null;
    var interval = intervals.insertChangingInterval(
      vsync: this,
      animationSettings: widget.animatedListAnimationSettings,
      index: from,
      changeCount: removeCount,
      removeItemBuilder: removeItemBuilder,
      onChangingCompleted: _onIntervalChangingCompleted,
      onCompleted: _onIntervalCompleted,
      onDisposed: onDisposed,
    );
    childManager.onUpdateOnNewInterval(interval);
    return interval;
  }

  /// Called when interval states change to begin new animations.
  void _startAnimations() {
    if (!_notifying && !isReordering) {
      intervals.startAnimations(widget.coordinateAnimations);
    }
  }

  /// Called when a removing interval is completed.
  void _onIntervalRemovingCompleted(AnimatedListInterval interval) {
    childManager.onUpdateOnIntervalRemovedToResizing(interval);
    _startAnimations();
  }

  /// Called when a resizing interval is completed.
  void _onIntervalResizingCompleted(AnimatedListInterval interval) {
    if (interval.insertCount == 0) {
      childManager.onUpdateOnIntervalResizedToDisposing(interval);
    } else {
      childManager.onUpdateOnIntervalResizedToInserting(interval);
    }
    _startAnimations();
  }

  /// Called when a inserting interval is completed.
  void _onIntervalInsertingCompleted(AnimatedListInterval interval) {
    childManager.onUpdateOnIntervalInsertedToDisposing(interval);
    _startAnimations();
  }

  /// Called when a changing interval is completed.
  void _onIntervalChangingCompleted(AnimatedListInterval interval) {
    childManager.onUpdateOnIntervalChangedToDisposing(interval);
    _startAnimations();
  }

  /// Called when an interval has completed its lifecycle.
  void _onIntervalCompleted(AnimatedListInterval interval) {
    _startAnimations();
  }

  // The inserting builder picks the item directly from the final changed list.
  Widget _defaultInsertItemBuilder(
      BuildContext context, int index, AnimatedListInterval interval) {
    index += interval.index;
    var i = intervals.indexOf(interval);
    assert(i >= 0);
    for (var j = 0; j < i; j++) {
      var prevInterval = intervals[j];
      index += prevInterval.adjustingItemCount;
    }
    return callBuildDelegateCallback(context, index, true);
  }

  /// Copied from [RenderSliverMultiBoxAdaptor.indexOf].
  /// An assertion was removed to accept the dummy child element.
  @override
  int indexOf(final RenderBox child) {
    assert(child != null);
    final childParentData = child.parentData as SliverMultiBoxAdaptorParentData;
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    // assert(childParentData.index != null);
    //""""""""""""""""""""""""""""""""""""""""""""""""""""""
    return childParentData.index;
  }

  @override
  AnimatedSliverMultiBoxAdaptorElement get childManager =>
      super.childManager as AnimatedSliverMultiBoxAdaptorElement;

  bool _inPaint = false;

  /// This method has been overridden for the needs of the reordering for two reasons:
  /// - do not paint the original dragged item: we keep it in the sliver because
  ///   we want it to be kept alive during all the reoder, but with a zero size;
  /// - paint the dragged item at then end of all the others children.
  @override
  RenderBox childAfter(RenderBox child) {
    if (_inPaint && isReordering) {
      if (child == childManager.reorderingDraggedChild) return null;
      var after = super.childAfter(child);
      if (after == _reorderingRemovedChild) {
        after = super.childAfter(after);
      } else {
        after ??= childManager.reorderingDraggedChild;
      }
      return after;
    }
    return super.childAfter(child);
  }

  @override
  void paint(PaintingContext context, Offset offset) {
    _inPaint = true;
    super.paint(context, offset);
    _inPaint = false;
  }

  bool _inPerformLayout = false;

  /// This method has been overridden to eventually measure the size of new resizing intervals and
  /// to eventually adjust the scroll offset.
  /// If required by [AnimatedSliverList.updateScrollableWhenResizing] attribute, calls the
  /// [_notifyScrollable] method to update the scroll listeners when a resizing interval is animating.
  /// In case of reordering, the dragged item will be layouted and a scroll up or down could be invoked
  /// when the latter is at the top or the bottoom of the viewport.
  @override
  void performLayout() {
    assert(!childManager.havePendingUpdates);

    _inPerformLayout = true;

    var resizing = false;
    final childConstraints = constraints.asBoxConstraints();
    for (var interval in intervals) {
      if (interval.isInResizingState) {
        if (!resizing) resizing = true;
        if (interval.fromSize == null || interval.toSize == null) {
          _measureIntervalSize(interval, childConstraints);
        }
      }
    }

    var offsetCorrection = _adjustScrollOffset();

    super.performLayout();

    if (isReordering) _computeNewReorderingOffset();

    // prevent many items from scrolling up out of the viewport when
    // resizing intervals placed before the first child shrink
    if (!isReordering) {
      offsetCorrection += geometry.scrollOffsetCorrection ?? 0.0;
      if (offsetCorrection != 0.0) {
        geometry = SliverGeometry(
          scrollExtent: geometry.scrollExtent,
          paintExtent: geometry.paintExtent,
          paintOrigin: geometry.paintOrigin,
          layoutExtent: geometry.layoutExtent,
          maxPaintExtent: geometry.maxPaintExtent,
          maxScrollObstructionExtent: geometry.maxScrollObstructionExtent,
          hitTestExtent: geometry.hitTestExtent,
          visible: geometry.visible,
          hasVisualOverflow: geometry.hasVisualOverflow,
          scrollOffsetCorrection: offsetCorrection,
          cacheExtent: geometry.cacheExtent,
        );
      }
    }

    if (resizing && !isReordering && widget.updateScrollableWhenResizing) {
      _notifyScrollable();
    }

    childManager.reorderingDraggedChild
        ?.layout(constraints.asBoxConstraints(), parentUsesSize: true);

    if (childManager.reorderingDraggedChild != null) {
      var currentOffset = currentReorderingChildOffset;
      (childManager.reorderingDraggedChild.parentData
              as SliverMultiBoxAdaptorParentData)
          .layoutOffset = currentOffset;

      // scroll up/down as needed while dragging
      var fromOffset = constraints.scrollOffset;
      var toOffset = fromOffset + constraints.remainingPaintExtent;

      var controller = Scrollable.of(childManager)?.widget?.controller;
      var position = controller?.position;

      var delta = 0.0;
      if (currentOffset < fromOffset && position.extentBefore > 0.0) {
        delta = -_kReorderingScrollSpeed;
      } else if (currentOffset + _reorderingChildSize > toOffset &&
          position.extentAfter > 0.0) {
        delta = _kReorderingScrollSpeed;
      }

      if (delta != 0.0) {
        final value = position.pixels + delta;
        WidgetsBinding.instance.addPostFrameCallback((_) {
          controller.jumpTo(value);
          markNeedsLayout();
        });
      }
    }

    _inPerformLayout = false;
  }

  RenderBox _findEarliestUsefulChild() {
    if (firstChild == null) {
      return null;
    }

    final scrollOffset = constraints.scrollOffset + constraints.cacheOrigin;
    assert(scrollOffset >= 0.0);
    final remainingExtent = constraints.remainingCacheExtent;
    assert(remainingExtent >= 0.0);

    var earliestUsefulChild = firstChild;

    if (childScrollOffset(firstChild) == null) {
      while (earliestUsefulChild != null &&
          childScrollOffset(earliestUsefulChild) == null) {
        earliestUsefulChild = childAfter(earliestUsefulChild);
      }
      if (firstChild == null) {
        return null;
      }
    }

    return earliestUsefulChild;
  }

  /// Updates the [Geometry] is necessary to correct the scroll offset taking into account
  /// of the space changed when resizing intervals from the last frame.
  /// The rendered items will also be moved up or down according to it.
  double _adjustScrollOffset() {
    var offsetCorrection = 0.0;

    var firstChild = _findEarliestUsefulChild();
    if (firstChild == null) return 0.0;

    final parentData =
        (firstChild.parentData as SliverMultiBoxAdaptorParentData);

    final firstIndex = parentData.index;

    // intervals.forEach((interval) {
    //   offsetCorrection += interval.markLayoutPerformed(firstIndex);
    // });
    offsetCorrection = childManager.calculateOffsetCorrection(firstIndex);

    if (offsetCorrection != 0.0) {
      final firstOffset = parentData.layoutOffset;
      if (firstOffset + offsetCorrection < 0.0) {
        offsetCorrection = -firstOffset; // bring back first offset to zero
      }
      parentData.layoutOffset += offsetCorrection;
    }

    return offsetCorrection;
  }

  /// Consider also the dragged child.
  @override
  void visitChildren(visitor) {
    super.visitChildren(visitor);
    if (childManager.reorderingDraggedChild != null) {
      assert(isReordering);
      visitor.call(childManager.reorderingDraggedChild);
    }
  }

  /// Dispatch a "fake" change to the [ScrollPosition] to force the listeners (like a [ScrollBar])
  /// to refresh its state.
  void _notifyScrollable() {
    var position = Scrollable.of(childManager)?.widget?.controller?.position;
    if (position == null) return;
    ScrollUpdateNotification(
            metrics: position.copyWith(),
            context: childManager,
            scrollDelta: 0,
            dragDetails: null)
        .dispatch(childManager);
  }

  /// Estimates the max scroll offset based on the rendered viewport data.
  double extrapolateMaxScrollOffset(
    final int firstIndex,
    final int lastIndex,
    final double leadingScrollOffset,
    final double trailingScrollOffset,
    final int childCount,
  ) {
    assert(!childManager.havePendingUpdates);

    if (lastIndex == childCount - 1) return trailingScrollOffset;

    var innerSpace = 0.0, trailingSpace = 0.0;
    var innerCount = 0, trailingCount = 0;
    for (final interval in intervals) {
      if (!interval.isInResizingState) continue;
      if (firstIndex <= interval.index) {
        if (interval.index <= lastIndex) {
          innerCount++;
          innerSpace += interval.currentSize;
        } else {
          trailingCount++;
          trailingSpace += interval.currentSize;
        }
      }
    }

    var ret = trailingScrollOffset + trailingSpace;

    final remainingCount = childCount - lastIndex - trailingCount - 1;
    if (remainingCount > 0) {
      var reifiedCount = 1 + lastIndex - firstIndex - innerCount;
      if (isReordering) {
        // exclude from refied count the zero sized removed/long-tapped child
        var i = _reorderingPickedIndex;
        for (final interval in intervals) {
          assert(interval.isInResizingState);
          if (interval.index > i) break;
          i++;
        }
        if (firstIndex <= i && i <= lastIndex) {
          reifiedCount--;
        }
      }
      double averageExtent;
      if (reifiedCount > 0) {
        averageExtent =
            (trailingScrollOffset - leadingScrollOffset - innerSpace) /
                reifiedCount;
      } else {
        var size = 0.0;
        var count = 0;
        for (final interval in intervals) {
          if (interval.isInResizingState) {
            if (firstIndex <= interval.index && interval.index <= lastIndex) {
              assert(interval.fromSize != null && interval.toSize != null);
              size += interval.fromSize + interval.toSize;
              count += interval.removeCount + interval.insertCount;
            }
          }
        }
        averageExtent = size / count;
      }
      ret += averageExtent * remainingCount;
    }
    return ret;
  }

  /// This method has been overridden in order to return the correct size of the resizing intervals.
  /// During reordering, the original of the dragged item must return a zero size.
  @override
  @protected
  double paintExtentOf(final RenderBox child) {
    var parentData = child.parentData as SliverMultiBoxAdaptorParentData;
    if (isReordering && child == _reorderingRemovedChild) return 0.0;
    if (parentData.index != null && _inPerformLayout) {
      var interval = childManager.intervalAtIndex(indexOf(child));
      if (interval != null && interval.isInResizingState) {
        return _measureIntervalSize(interval, constraints.asBoxConstraints());
      }
    }
    return super.paintExtentOf(child);
  }

  /// Calculate the space required by the specified resizing interval.
  /// If the range has been measured already, the previously calculated value is returned.
  double _measureIntervalSize(
      AnimatedListInterval interval, BoxConstraints childConstraints) {
    assert(interval != null && interval.isInResizingState);
    double calculate(int n, NullableIndexedWidgetBuilder builder) {
      var size = 0.0;
      int i;
      final maxSize = constraints.viewportMainAxisExtent +
          (parent as RenderViewport).cacheExtent * 2;
      invokeLayoutCallback<SliverConstraints>((SliverConstraints constraints) {
        for (i = 0; i < n; i++) {
          if (size > maxSize) break;
          final w = builder.call(childManager, i);
          final r = childManager.layoutDummyChild(w, childConstraints);
          size += paintExtentOf(r);
        }
        if (i < n) size *= (n / i);
        childManager.dropDummyChild();
      });
      return size;
    }

    interval.measureSizesIfNeeded(
        () => calculate(interval.removeCount, interval.removeItemBuilder),
        () => calculate(interval.insertCount,
            (c, i) => _defaultInsertItemBuilder(c, i, interval)));
    assert(interval.fromSize != null && interval.toSize != null);
    return interval.currentSize;
  }

  RenderBox _reorderingRemovedChild;
  int _reorderingPickedIndex, _reorderingDropIndex;
  double _reorderingChildSize;
  double _reorderingStartOffset,
      _reorderingDeltaOffset,
      _reorderingLastOffset,
      _reorderingScrollOffset;

  bool get isReordering => _reorderingPickedIndex != null;

  bool get isAnimating => intervals.isNotEmpty;

  double get currentReorderingChildOffset {
    return isReordering
        ? min(
            max(
                0,
                _reorderingStartOffset +
                    _reorderingDeltaOffset +
                    (constraints.scrollOffset - _reorderingScrollOffset)),
            geometry.maxPaintExtent - _reorderingChildSize)
        : 0.0;
  }

  void _startReordering(int index, double offsetX, double offsetY) {
    _reorderingRemovedChild = childManager.onUpdateOnStartReording(index);

    _reorderingChildSize = paintExtentOf(_reorderingRemovedChild);

    _reorderingScrollOffset = constraints.scrollOffset;
    _reorderingPickedIndex = index;
    _reorderingDropIndex = index;
    _reorderingDeltaOffset = 0;
    _reorderingStartOffset = _reorderingLastOffset =
        (_reorderingRemovedChild.parentData as SliverMultiBoxAdaptorParentData)
            .layoutOffset;

    var interval = intervals.insertReorderingInterval(
      vsync: this,
      animationSettings: widget.animatedListAnimationSettings,
      index: index + 1,
      size: _reorderingChildSize,
      appearing: false,
      onResizingCompleted: _onIntervalResizingCompleted,
    );
    childManager.onUpdateOnNewInterval(interval);
  }

  void _updateReordering(double dy) {
    assert(isReordering);
    _reorderingDeltaOffset = dy;
    markNeedsLayout();
  }

  void _computeNewReorderingOffset() {
    assert(!childManager.havePendingUpdates);

    int newReorderingIndex;

    var currentOffset = currentReorderingChildOffset;

    if (currentOffset == _reorderingLastOffset) return;

    var y1 = currentOffset;
    var y2 = y1 + _reorderingChildSize;

    if (currentOffset > _reorderingLastOffset) {
      // move down
      outer:
      for (var child = firstChild; child != null; child = childAfter(child)) {
        if (isBoundaryChild(child)) continue;
        var space = 0.0;
        var adj = 0;
        var j = (child.parentData as SliverMultiBoxAdaptorParentData).index;
        for (var i in intervals) {
          if (i.index == j) {
            continue outer;
          } else if (i.index > j) break;
          space += i.currentSize;
          adj++;
        }
        var pd = child.parentData as SliverMultiBoxAdaptorParentData;
        var yc1 = pd.layoutOffset - space + _reorderingChildSize;
        var yc2 = yc1 + paintExtentOf(child);

        if (y2 < yc1 + (yc2 - yc1) * 0.75) break;

        newReorderingIndex = j - adj + 1;
      }
    } else {
      // move up
      outer:
      for (var child = lastChild; child != null; child = childBefore(child)) {
        if (isBoundaryChild(child)) continue;
        var space = 0.0;
        var adj = 0;
        var j = (child.parentData as SliverMultiBoxAdaptorParentData).index;
        for (var i in intervals) {
          if (i.index == j) {
            continue outer;
          } else if (i.index > j) break;
          space += i.currentSize;
          adj++;
        }
        var pd = child.parentData as SliverMultiBoxAdaptorParentData;
        var yc1 = pd.layoutOffset - space;
        var yc2 = yc1 + paintExtentOf(child);

        if (y1 > yc1 + (yc2 - yc1) * 0.25) {
          break;
        }

        newReorderingIndex = j - adj;
      }
    }

    if (newReorderingIndex != null) {
      if (newReorderingIndex >= _reorderingPickedIndex + 1) {
        newReorderingIndex--;
      }
      WidgetsBinding.instance.addPostFrameCallback((_) {
        _updateReorderingIntervals(newReorderingIndex);
      });
    }

    _reorderingLastOffset = currentOffset;
  }

  bool isBoundaryChild(RenderBox child) => childManager.widgetOf(
          (child.parentData as SliverMultiBoxAdaptorParentData)?.index)
      is _BoundaryChild;

  void _updateReorderingIntervals(int newIndex) {
    if (_reorderingDropIndex == newIndex ||
        !(widget.delegate.onReorderSwap
                ?.call(_reorderingPickedIndex, newIndex) ??
            false)) {
      return;
    }

    _reorderingDropIndex = newIndex;

    if (newIndex > _reorderingPickedIndex) newIndex++;

    var adj = 0, j = 0;
    var create = true;

    for (var i = 0; i < intervals.length; i++) {
      var interval = intervals[i];

      assert(interval.isInResizingState && interval.isReordering);
      if (interval.index != newIndex + adj) {
        if (interval.isWaiting) {
          // the interval is in its full size waiting to be closed... so let's give it the signal
          interval.startAnimation();
        } else {
          // the interval is animating to reach its full size... let's give the signal to resize to dismiss
          if (interval.resize(0.0)) {
            i--;
            continue;
          }
        }
      } else {
        // the interval already exists, just resize it to its full size
        interval.resize(_reorderingChildSize);
        create = false;
      }
      if (interval.index < newIndex + adj) j++;
      adj++;
    }

    if (create) {
      var interval = intervals.insertReorderingInterval(
        vsync: this,
        animationSettings: widget.animatedListAnimationSettings,
        index: newIndex + j,
        size: _reorderingChildSize,
        appearing: true,
        onResizingCompleted: _onIntervalResizingCompleted,
      );
      childManager.onUpdateOnNewInterval(interval);
    }
  }

  void _stopReordering(bool cancel) {
    intervals.finishReorder();

    childManager.onUpdateOnStopReording(_reorderingPickedIndex,
        cancel ? _reorderingPickedIndex : _reorderingDropIndex);

    _clearReordering();
  }

  void _clearReordering() {
    _reorderingRemovedChild = null;
    _reorderingPickedIndex = null;
    _reorderingDropIndex = null;
    _reorderingChildSize = null;
    _reorderingStartOffset = null;
    _reorderingDeltaOffset = null;
    _reorderingLastOffset = null;
    _reorderingScrollOffset = null;
  }

  void didChildManagerMounted() {
    widget.controller?._setAdaptor(this);
  }

  void didChildManagerUnmounted() {
    widget.controller?._setAdaptor(null);
  }

  /// Returns true if the item at the [index] position of the sliver is an animating item.
  bool isIndexAnimated(int index) {
    for (var j = 0; j < intervals.length; j++) {
      var interval = intervals[j];

      switch (interval.state) {
        case AnimatedListIntervalState.REMOVING:
          if (index >= interval.index) {
            if (index < (interval.index + interval.removeCount)) {
              return true;
            }
          }
          break;
        case AnimatedListIntervalState.CHANGING:
          if (index >= interval.index) {
            if (index < (interval.index + interval.removeCount)) {
              return true;
            }
          }
          break;
        case AnimatedListIntervalState.RESIZING:
          if (index == interval.index) {
            return true;
          } else if (index > interval.index) {}
          break;
        case AnimatedListIntervalState.INSERTING:
          if (index >= interval.index &&
              index < (interval.index + interval.insertCount)) {
            return true;
          }
          break;
        default:
          throw 'Illegal state';
      }
    }
    return false;
  }

  String toDebugString() => intervals.toDebugString();
}

// Copied from standard package.
class _WidgetTicker extends Ticker {
  _WidgetTicker(TickerCallback onTick, this._creator) : super(onTick);

  final AnimatedRenderSliverList _creator;

  @override
  void dispose() {
    _creator._removeTicker(this);
    super.dispose();
  }
}

//---------------------------------------------------------------------------------------------
// AnimatedSliverChildBuilderDelegate
//---------------------------------------------------------------------------------------------

int _kDefaultSemanticIndexCallback(Widget _, int localIndex) => localIndex;

int _kChildCount() => null;

/// Class copied from [SliverChildBuilderDelegate].
///
/// The [builder] has been changed using a new callback that has a new 'animating' parameter
/// indicating when to build the item in an animating state or not. For example, in animating
/// state you should avoid to use a key because a duplicate key exception will arise.
///
/// The [childCount] attribute has been replaced with a callback function. This callback
/// has to return the actual length of your list that will be changed after every change,
/// insertion, removal and replacement. A null value can be also returned, especially for
/// infinite lists.
///
/// New redorder callbacks have been added in order to handle the reordering:
/// - [onReorderStart] called first when it would like to start reordering the specified item by index;
///   two other additional parameters are passed that indicate the offset of the tap/click
///   from the top left corner of the item; if false is returned, the reorder won't start at all;
/// - [onReorderFeedback] called whenever the item is moved by dragging; the item's index is
///   passed again, then the offset from the top of the sliver and finally a delta indicating
///   the distance of the new tapped/clicked point from the original one;
/// - [onReorderSwap] called when the dragged item, indicated by the first index, would like to
///   swap with the item indicated by the second index; if false is returned, the swap
///   won't be proposed at all.
/// - [onReorderComplete] called at the end of reorder when the swap must actually be done;
///   implement this callback to update your underlying list; return false if you want
///   to cancel the swap.
class AnimatedSliverChildBuilderDelegate extends SliverChildDelegate {
  const AnimatedSliverChildBuilderDelegate(
    this.builder, {
    this.findChildIndexCallback,
    this.childCount = _kChildCount,
    this.addAutomaticKeepAlives = true,
    this.addRepaintBoundaries = true,
    this.addSemanticIndexes = true,
    this.semanticIndexCallback = _kDefaultSemanticIndexCallback,
    this.semanticIndexOffset = 0,
    this.onReorderStart,
    this.onReorderFeedback,
    this.onReorderSwap,
    this.onReorderComplete,
  })  : assert(builder != null),
        assert(addAutomaticKeepAlives != null),
        assert(addRepaintBoundaries != null),
        assert(addSemanticIndexes != null),
        assert(semanticIndexCallback != null);

  final AnimatedNullableIndexedWidgetBuilder builder;
  final int Function() childCount;
  final bool addAutomaticKeepAlives;
  final bool addRepaintBoundaries;
  final bool addSemanticIndexes;
  final int semanticIndexOffset;
  final SemanticIndexCallback semanticIndexCallback;
  final ChildIndexGetter findChildIndexCallback;

  final bool Function(int index, double dx, double dy) onReorderStart;
  final void Function(int index, double offset, double dy) onReorderFeedback;
  final bool Function(int index, int dropIndex) onReorderSwap;
  final bool Function(int index, int dropIndex) onReorderComplete;

  /// Not used anymore.
  @override
  @mustCallSuper
  int findIndexByKey(Key key) {
    throw "The method 'findIndexByKey' is not used anymore";
  }

  /// Copied from [SliverChildBuilderDelegate.build]. This method has been modified to
  /// handle the new animating attribute that indicates when an item should be build
  /// in an animating way or not. In addition, the new callback
  /// [AnimatedSliverChildBuilderDelegate.childCount] will be called instead of
  /// just an attribute.
  /// If the reordering is enabled, a long tap detector will also be added.
  @override
  Widget build(BuildContext context, int index) {
    assert(builder != null);
    final childManager = context as AnimatedSliverMultiBoxAdaptorElement;
    final animating = childManager?.animating ?? true;
    final count = childCount?.call();
    if (index < 0 || (count != null && index >= count)) return null;
    Widget child;
    try {
      child = builder(context, index, animating);
    } catch (exception, stackTrace) {
      child = _createErrorWidget(exception, stackTrace);
    }
    if (child == null) {
      return null;
    }

    final Key key = child.key != null ? _SaltedValueKey(child.key) : null;
    if (addRepaintBoundaries) child = RepaintBoundary(child: child);
    if (addSemanticIndexes && !animating) {
      final semanticIndex = semanticIndexCallback(child, index);
      if (semanticIndex != null) {
        child = IndexedSemantics(
            index: semanticIndex + semanticIndexOffset, child: child);
      }
    }

    if (childManager.widget.reorderable) {
      child = GestureDetector(
        child: child,
        onLongPressStart: (d) {
          var renderObject = childManager.renderObject;
          if (!renderObject.isAnimating &&
              (onReorderStart?.call(
                      index, d.localPosition.dx, d.localPosition.dy) ??
                  false)) {
            renderObject._startReordering(index, 0, 0);
          }
        },
        onLongPressEnd: (d) {
          var ro = childManager.renderObject;
          if (ro.isReordering) {
            var cancel = (ro._reorderingPickedIndex ==
                    ro._reorderingDropIndex) ||
                (!(onReorderComplete?.call(
                        ro._reorderingPickedIndex, ro._reorderingDropIndex) ??
                    false));
            ro._stopReordering(cancel);
          }
        },
        onLongPressMoveUpdate: (d) {
          var ro = childManager.renderObject;
          if (ro.isReordering) {
            ro._updateReordering(d.localOffsetFromOrigin.dy);
            onReorderFeedback?.call(ro._reorderingPickedIndex,
                ro.currentReorderingChildOffset, ro._reorderingDeltaOffset);
          }
        },
      );
    }

    if (addAutomaticKeepAlives && !animating) {
      child = AutomaticKeepAlive(child: child);
    }

    return KeyedSubtree(child: child, key: key);
  }

  /// Method not supported yet. Don't override it!
  @override
  double estimateMaxScrollOffset(
    int firstIndex,
    int lastIndex,
    double leadingScrollOffset,
    double trailingScrollOffset,
  ) {
    throw "'estimateMaxScrollOffset' method not supported yet.";
  }

  /// Copied from [SliverChildBuilderDelegate.estimatedChildCount]. The new callback
  /// [AnimatedSliverChildBuilderDelegate.childCount] will be called instead
  /// of just an attribute.
  @override
  int get estimatedChildCount => childCount();

  /// Copied from [SliverChildBuilderDelegate.shouldRebuild].
  @override
  bool shouldRebuild(
          covariant AnimatedSliverChildBuilderDelegate oldDelegate) =>
      true;
}

class _SaltedValueKey extends ValueKey<Key> {
  const _SaltedValueKey(Key key)
      : assert(key != null),
        super(key);
}

NullableIndexedWidgetBuilder _shiftItemBuilder(
    final NullableIndexedWidgetBuilder builder, final int leading) {
  if (leading == 0 || builder == null) return builder;
  return (context, index) => builder.call(context, index + leading);
}

// Return a Widget for the given Exception (copied from standard package).
Widget _createErrorWidget(Object exception, StackTrace stackTrace) {
  final details = FlutterErrorDetails(
    exception: exception,
    stack: stackTrace,
    library: 'widgets library',
    context: ErrorDescription('building'),
  );
  FlutterError.reportError(details);
  return ErrorWidget.builder(details);
}

class _BoundaryChild extends SizedBox {
  _BoundaryChild() : super.shrink();
}
