import 'dart:math';

import 'package:flutter/widgets.dart';
import 'animated_sliver_list.dart';
import 'circular_list.dart';

const int _kWindowSize = 30;

bool _kEquals(dynamic a, dynamic b) => a == b;

/// Adapter that takes a tree model as input and transforms it into a linear list view.
/// You must specify the data type [T] of its nodes.
class TreeListAdapter<T> {
  final T Function(T node) parentOf;
  final int Function(T node) childrenCount;
  final T Function(T node, int index) childAt;
  final bool Function(T node) isNodeExpanded;
  final int Function(T parent, T node) indexOfChild;
  final bool Function(T nodeA, T nodeB) equals;

  // the list never contain the root node, although it is shown
  final CircularList<T> _list;

  final int windowSize;
  final bool includeRoot;
  final int startingLevel;

  T root;
  int _offset = 0, _count;

  int get _endOffset => _offset + windowSize;

  /// Creates a new tree adapter.
  /// A [root] node has to be specified.
  ///
  /// The model consists of following callbacks:
  /// - [parentOf] returns the parent node of a node;
  /// - [childrenCount] returns the number of the children of a node;
  /// - [childAt] returns the child node of a parent node at a specific position;
  /// - [isNodeExpanded] returns true if the node is expanded, false if it is collapsed;
  /// - [indexOfChild] returns the position of a child node;
  /// - [equals] returns true if two nodes are equals; if this callback is omitted, the ==
  ///   operator will be used;
  ///
  /// If [includeRoot] is set to true, the list view will also show the root node as the first item.
  ///
  /// You can specify by [windowSize] the size of the circular list used to cache the items.
  ///
  /// You can also pass by [initialCount] the initial count of all tree nodes if it is known,
  /// to prevent the full scan of the tree in order to calculate its size.
  TreeListAdapter({
    @required this.root,
    @required this.parentOf,
    @required this.childrenCount,
    @required this.childAt,
    @required this.isNodeExpanded,
    @required this.indexOfChild,
    this.equals = _kEquals,
    this.includeRoot = true,
    this.windowSize = _kWindowSize,
    int initialCount,
    this.startingLevel = 0,
  })  : assert(root != null),
        assert(parentOf != null),
        assert(childrenCount != null),
        assert(childAt != null),
        assert(isNodeExpanded != null),
        assert(indexOfChild != null),
        assert(equals != null),
        assert(includeRoot != null),
        assert(windowSize != null),
        assert(startingLevel != null),
        _list = CircularList(windowSize) {
    if (initialCount != null) {
      assert(initialCount == count);
      _count = initialCount;
    }
  }

  /// Returns a sub tree starting from the [node].
  ///
  /// If [includeRoot] is set to true, the root [node] will be considered as the first item.
  ///
  /// If [alwaysExpandRoot] is set to true, the root [node] is considered expanded despite the callback
  /// [isNodeExpanded] callback.
  ///
  /// If [keepCurrentLevel] is set to true, the [node] will keep its current level also in the subtree.
  ///
  /// You can assign a new window size for the subtree specifying the attribute [windowSize]. If you set
  /// this to null, the subtree will inherit the same [windowSize] of this original tree.
  TreeListAdapter<T> subTreeOf(T node,
      [bool includeRoot = false,
      bool alwaysExpandRoot = true,
      bool keepCurrentLevel = true,
      int windowSize = _kWindowSize]) {
    final _this = this;
    return TreeListAdapter(
      root: node,
      parentOf: (n) => (equals(n, node)) ? null : _this.parentOf(n),
      childrenCount: childrenCount,
      childAt: childAt,
      isNodeExpanded: (n) => (alwaysExpandRoot && equals(n, node))
          ? true
          : _this.isNodeExpanded(n),
      indexOfChild: indexOfChild,
      equals: equals,
      includeRoot: includeRoot,
      windowSize: windowSize ?? this.windowSize,
      // initialCount: null,
      startingLevel:
          keepCurrentLevel ? (levelOf(node) + (includeRoot ? 0 : 1)) : 0,
    );
  }

  /// Returns the number of items required to show in a list view the entire tree.
  int get count {
    if (_count == null) {
      _count = _countSubNodesOf(root);
      if (includeRoot) _count++;
    }
    return _count;
  }

  /// Count how many items are required to show this node and its children in a list view,
  /// taking into account when a node is collapsed or not.
  int countSubTree(T node) {
    var n = 1;
    if (isNodeExpanded(node)) n += _countSubNodesOf(node);
    return n;
  }

  /// Count how many items are required to show the children of this node in a list view.
  /// If [treatAsAllExpanded] is set to true, descendants will be treated
  /// as if they were always expanded, otherwise the current expand status will be
  /// considered.
  int _countSubNodesOf(T node, [bool treatAsAllExpanded = false]) {
    var n = childrenCount(node);
    var r = n;
    for (var i = 0; i < n; i++) {
      var child = childAt(node, i);
      if (isNodeExpanded(child) || treatAsAllExpanded) {
        r += _countSubNodesOf(child);
      }
    }
    return r;
  }

  /// Returns true if the [node] is the [root] node. The calback [equals] will be used.
  bool isRootNode(T node) => equals(node, root);

  /// Returns true if the [node] is a leaf node, that is without children.
  bool isLeaf(T node) => childrenCount(node) == 0;

  /// Calculates the level of hierarchy of the [node], taking into account the [startingLevel].
  /// You can use this value to shift the widget's position right.
  int levelOf(T node) {
    assert(node != null);
    var level = includeRoot ? startingLevel : startingLevel - 1;
    for (var n = node; !isRootNode(n); n = parentOf(n)) {
      level++;
    }
    return level;
  }

  int _ensureIndex(T node, int index) {
    assert(node != null);
    assert(index == null || nodeToIndex(node) == index);
    index ??= nodeToIndex(node);
    return index;
  }

  Range _notify(
    int from,
    int to,
    T node,
    void Function() fn,
    int index,
    AnimatedListController controller,
    void Function(int from, int to) cb,
    void Function(int from, int count) controllerCb,
  ) {
    assert(from != null && to != null);
    if (index == null) return null;

    cb?.call(from, to);

    final count = to - from;
    if (includeRoot) from++;

    fn?.call();

    if (controller != null) {
      controllerCb?.call(from, count);
      return null;
    } else {
      return Range(from, count);
    }
  }

  /// This method has to be called when the [node] is about to be expanded.
  /// Pass your function to [expandFn] that takes care of actually expanding the node.
  /// If you also have the index of the corresponding list view item, that's better
  /// pass it through the [index] attribute, otherwise you can just omit it.
  /// If you pass a [controller], the linked animated list view will be automatically notified.
  /// If you don't pass a [controller], a [Range] will be returned to indicate
  /// the range of items of the list view involved in the modification.
  Range notifyNodeExpanding(T node, void Function() expandFn,
      {int index, AnimatedListController controller}) {
    index = _ensureIndex(node, index);
    final from = index + (includeRoot ? 0 : 1); // first child index
    final to = from + _countSubNodesOf(node);
    return _notify(from, to, node, expandFn, index, controller, _insert,
        (from, count) {
      controller.notifyInsertedRange(from, count);
    });
  }

  /// This method has to be called when a new subtree [newSubTree] is about to be inserted
  /// as child of the [parentNode] at the position [position].
  /// Pass your function to [insertFn] that takes care of actually inserting the node.
  /// If you also have the parent's index of the corresponding list view item, that's better
  /// pass it through the [index] attribute, otherwise you can just omit it.
  /// If you pass a [controller], the linked animated list view will be automatically notified.
  /// If you don't pass a [controller], a [Range] will be returned to indicate
  /// the range of items of the list view involved in the modification.
  Range notifyNodeInserting(
      T newSubTree, T parentNode, int position, void Function() insertFn,
      {int index, AnimatedListController controller}) {
    if (!isNodeExpanded(parentNode)) {
      insertFn.call();
      return null;
    }
    index = _ensureIndex(parentNode, index);
    assert(newSubTree != null && position != null);
    assert(position >= 0 && position <= childrenCount(parentNode));
    var from = index + (includeRoot ? 0 : 1); // first child index
    for (var i = 0; i < position; i++) {
      from += countSubTree(childAt(parentNode, i));
    }
    final to = from + countSubTree(newSubTree);
    return _notify(from, to, parentNode, insertFn, index, controller, _insert,
        (from, count) {
      controller.notifyInsertedRange(from, count);
    });
  }

  /// This method has to be called when the [node] is about to be collapsed.
  /// Pass your function to [collapseFn] that takes care of actually collapsing the node.
  /// If you also have the index of the corresponding list view item, that's better
  /// pass it through the [index] attribute, otherwise you can just omit it.
  /// If you pass a [controller], the linked animated list view will be automatically notified.
  /// In this case you have to indicate a [builder] to build the collapsed subtree.
  /// If you don't pass a [controller], a [Range] will be returned to indicate
  /// the range of items of the list view involved in the modification.
  Range notifyNodeCollapsing(T node, void Function() collapseFn,
      {int index,
      AnimatedListController controller,
      TreeItemBuilder<T> builder}) {
    index = _ensureIndex(node, index);
    final from = index + (includeRoot ? 0 : 1); // first child index
    final to = from + _countSubNodesOf(node);
    return _notify(from, to, node, collapseFn, index, controller, _remove,
        (from, count) {
      assert(builder != null);
      final subAdapter = subTreeOf(node);
      controller.notifyRemovedRange(from, count,
          (context, idx) => builder.call(subAdapter, context, idx, true));
    });
  }

  /// This method has to be called when the [node] is about to be removed.
  /// Pass your function to [removeFn] that takes care of actually removing the node.
  /// If you also have the index of the corresponding list view item, that's better
  /// pass it through the [index] attribute, otherwise you can just omit it.
  /// If you pass a [controller], the linked animated list view will be automatically notified.
  /// In this case you have to indicate a [builder] to build the removed subtree.
  /// If you don't pass a [controller], a [Range] will be returned to indicate
  /// the range of items of the list view involved in the modification.
  Range notifyNodeRemoving(T node, void Function() removeFn,
      {int index,
      AnimatedListController controller,
      TreeItemBuilder<T> builder}) {
    assert(!equals(node, root)); // root cannot be removed!
    index = _ensureIndex(node, index);
    final from = index - (includeRoot ? 1 : 0);
    final to = from + countSubTree(node);
    return _notify(from, to, node, removeFn, index, controller, _remove,
        (from, count) {
      assert(builder != null);
      final subAdapter = subTreeOf(node, true);
      controller.notifyRemovedRange(from, count,
          (context, idx) => builder.call(subAdapter, context, idx, true));
    });
  }

  /// Call this method when the tree has been completly changed.
  void notifyTreeChanged(T root, [int initialCount]) {
    assert(root != null);
    _list.clear();
    this.root = root;
    _count = null;
    _offset = 0;
    if (initialCount != null) {
      assert(initialCount == count);
      _count = initialCount;
    }
  }

  //

  void _remove(int from, int to /* excluded */) {
    assert(from != null && to != null);
    assert(from <= to);
    if (from == to) return;
    _count -= to - from;
    if (from >= _endOffset || to <= _offset) return;
    if (from <= _offset && to >= _endOffset) {
      _list.clear();
    } else {
      _list.remove(max(from, _offset) - _offset, min(to, _endOffset) - _offset);
    }
  }

  void _insert(int from, int to /* excluded */) {
    assert(from != null && to != null);
    assert(from <= to);
    if (from == to) return;
    _count += to - from;
    if (from >= _endOffset) return;
    if (from <= _offset && to >= _endOffset) {
      _list.clear();
    } else {
      _list.insert(max(from, _offset) - _offset, min(to, _endOffset) - _offset);
    }
  }

  T _cachedNode(int index) =>
      (index < _offset || index >= _endOffset) ? null : _list[index - _offset];

  void _cacheNode(int index, T node) {
    if (index < _offset) {
      _list.shift(index - _offset);
      _offset = index;
    } else if (index >= _endOffset) {
      _list.shift(index - _endOffset + 1);
      _offset = index - windowSize + 1;
    }
    _list[index - _offset] = node;
  }

  /// Returns the index of the corresponding list view item for the [node].
  int nodeToIndex(T node) {
    if (equals(node, root)) return (includeRoot ? 0 : -1);
    for (var i = 0; i < windowSize; i++) {
      if (_list[i] != null && equals(_list[i], node)) {
        return i + _offset + (includeRoot ? 1 : 0);
      }
    }
    var i = _searchNode(node, root, includeRoot ? 0 : -1);
    if (i < 0) return -i - 1;
    return null;
  }

  int _searchNode(T nodeToSearch, T node, int index) {
    if (equals(nodeToSearch, node)) return -index - 1;
    if (!isNodeExpanded(node)) return index;
    var n = childrenCount(node);
    for (var i = 0; i < n; i++) {
      var child = childAt(node, i);
      index = _searchNode(nodeToSearch, child, ++index);
      if (index < 0) return index;
    }
    return index;
  }

  /// Returns the node underlying the list view item at position [index].
  T indexToNode(int index) {
    if (includeRoot) {
      if (index == 0) return root;
      index--;
    }
    var node = _cachedNode(index);
    if (node != null) return node;
    var i = index, j = index;
    bool next;
    do {
      next = false;
      if (--i >= _offset) {
        next = true;
        node = _cachedNode(i);
        if (node != null) {
          _iterateForward(node, i, index);
          node = _cachedNode(index);
          assert(node != null);
          return node;
        }
      }
      if (++j < _endOffset) {
        next = true;
        node = _cachedNode(j);
        if (node != null) {
          _iterateBackward(node, j, index);
          node = _cachedNode(index);
          assert(node != null);
          return node;
        }
      }
    } while (next);
    _iterateForward(root, -1, index);
    node = _cachedNode(index);
    assert(node != null);
    return node;
  }

  void _iterateForward(T node, int index, int toIndex) {
    index = _iterateForwardDescendants(node, index, toIndex);
    if (index >= toIndex) return;
    while (index < toIndex) {
      var parent = parentOf(node);
      var n = childrenCount(parent);
      var j = indexOfChild(parent, node);
      assert(j >= 0 && j < n);
      for (j++; j < n; j++) {
        var child = childAt(parent, j);
        _cacheNode(++index, child);
        if (index >= toIndex) return;
        index = _iterateForwardDescendants(child, index, toIndex);
        if (index >= toIndex) return;
      }
      node = parent;
    }
  }

  int _iterateForwardDescendants(T node, int index, int toIndex) {
    if (index < toIndex && isNodeExpanded(node)) {
      var n = childrenCount(node);
      for (var i = 0; i < n; i++) {
        var child = childAt(node, i);
        _cacheNode(++index, child);
        if (index >= toIndex) break;
        index = _iterateForwardDescendants(child, index, toIndex);
        if (index >= toIndex) break;
      }
    }
    return index;
  }

  void _iterateBackward(T node, int index, int toIndex) {
    while (index > toIndex) {
      var parent = parentOf(node);
      var n = childrenCount(parent);
      var j = indexOfChild(parent, node);
      assert(j >= 0 && j < n);
      for (j--; j >= 0; j--) {
        var child = childAt(parent, j);
        index = _iterateBackwardDescendants(child, index, toIndex);
        if (index <= toIndex) return;
      }
      _cacheNode(--index, parent);
      node = parent;
    }
  }

  int _iterateBackwardDescendants(T node, int index, int toIndex) {
    if (isNodeExpanded(node)) {
      var n = childrenCount(node);
      for (var i = n - 1; i >= 0; i--) {
        var child = childAt(node, i);
        index = _iterateBackwardDescendants(child, index, toIndex);
        if (index <= toIndex) return index;
      }
    }
    _cacheNode(--index, node);
    return index;
  }

  T lastChildrenOf(T node) => childAt(node, childrenCount(node) - 1);

  bool descendsFrom(T parentNode, T node) {
    for (var n = parentNode; !isRootNode(n); n = parentOf(n)) {
      if (node == n) return true;
    }
    return false;
  }
}

class Range {
  final int from, length;
  int get to => from + length;
  Range(this.from, this.length);
}

typedef TreeItemBuilder<T> = Widget Function(
    TreeListAdapter<T> adapter, BuildContext context, int idx, bool animating);
